"""This module contains functions for LS2 data analysis that utilize
the multiprocessing Python library for speeding up the analysis
by using all available CPU cores instead of a single one.

This module can be imported with the following functions:

* calculate_and_save_timestamp_differences_mp - unpacks the binary data,
    calculates timestamp differences, and saves into a '.feather' file.
    Works with firmware versions '2208' and '2212b'. The multiprocessing
    version that utilizes all available CPU cores.
    
* calculate_and_save_timestamp_differences_full_sensor_mp - unpacks the
    binary data, calculates timestamp differences and saves into a
    '.feather' file. Works with firmware versions '2208', '2212s' and
    '2212b'. Analyzes data from both sensor halves/both FPGAs. Useful
    for data where the signal is at 0 across the whole sensor from the
    start of data collecting. The multiprocessing version that
    utilizies all CPU cores available.

* compact_share_mp - unpacks all '.dat' files in the given folder,
    collects the number of timestamps in each pixel and packs it into a
    '.txt' file, calculates timestamp differences and packs them into a 
    '.feather' file. The multiprocessing version that utilizes all
    available CPU cores.

"""

import functools
import glob
import multiprocessing
import os
import sys
from dataclasses import dataclass
from zipfile import ZipFile

import numpy as np
import pandas as pd
from pyarrow import feather as ft

from LinoSPAD2.functions import calc_diff as cd
from LinoSPAD2.functions import unpack as f_up
from LinoSPAD2.functions import utils

# from tqdm import tqdm


@dataclass
class DataParamsConfig:
    """Configuration parameters for timestamp differences calculation.

    Parameters
    ----------
    pixels : list
        List of pixel numbers for which the timestamp differences should
        be calculated and saved or list of two lists with pixel numbers
        for peak vs. peak calculations.
    path : str
        Path to data files.
    daughterboard_number : str
        LinoSPAD2 daughterboard number.
    motherboard_number : str
        LinoSPAD2 motherboard (FPGA) number, including the "#".
    motherboard_number1 : str
        First LinoSPAD2 motherboard (FPGA) number, including the "#".
        Used for full sensor data analysis.
    motherboard_number2 : str
        Second LinoSPAD2 motherboard (FPGA) number. Used for full sensor
        data analysis.
    firmware_version : str
        LinoSPAD2 firmware version. Versions "2212s" (skip) and "2212b"
        (block) are recognized.
    timestamps : int, optional
        Number of timestamps per acquisition cycle per pixel. The default
        is 512.
    delta_window : float, optional
        Size of a window to which timestamp differences are compared.
        Differences in that window are saved. The default is 50e3 (50 ns).
    app_mask : bool, optional
        Switch for applying the mask for hot pixels. The default is True.
    include_offset : bool, optional
        Switch for applying offset calibration. The default is True.
    apply_calibration : bool, optional
        Switch for applying TDC and offset calibration. If set to 'True'
        while apply_offset_calibration is set to 'False', only the TDC
        calibration is applied. The default is True.
    absolute_timestamps : bool, optional
        Indicator for data with absolute timestamps. The default is
        False.
    """

    pixels: list
    path: str = ""
    daughterboard_number: str = ""
    motherboard_number: str = ""
    motherboard_number1: str = ""
    motherboard_number2: str = ""
    firmware_version: str = ""
    timestamps: int = 512
    delta_window: float = 50e3
    app_mask: bool = True
    include_offset: bool = True
    apply_calibration: bool = True
    absolute_timestamps: bool = False


def _calculate_timestamps_differences(
    file: str,
    result_queue: multiprocessing.Queue,
    data_params: DataParamsConfig,
) -> None:
    """Unpack data and collect timestamps differences for single file.

    Unpack the '.dat' data file, collect timestamps differences in the
    given window ("delta_window") and add them to the shared
    interprocess queue for further manipulations.

    Parameters
    ----------
    file : str
        Data file name, relative path.
    result_queue : multiprocessing.Queue
        Interprocess queue created using the multiprocessing.Queue method.
    data_params : DataParamsConfig
        An instance of the DataParamsConfig dataclass containing various
        configuration parameters for data processing.

    Returns
    -------
    None.
    """

    if data_params.firmware_version == "2212s":
        pix_coor = np.arange(256).reshape(4, 64).T
    elif data_params.firmware_version == "2212b":
        pix_coor = np.arange(256).reshape(64, 4)
    else:
        print("\nFirmware version is not recognized.")
        sys.exit()

    if data_params.app_mask is True:
        mask = utils.apply_mask(
            data_params.daughterboard_number, data_params.motherboard_number
        )
        if isinstance(data_params.pixels[0], int) and isinstance(
            data_params.pixels[1], int
        ):
            pixels = [pix for pix in data_params.pixels if pix not in mask]
        else:
            pixels = [pix for pix in data_params.pixels[0] if pix not in mask]
            pixels.extend(
                pix for pix in data_params.pixels[1] if pix not in mask
            )

    deltas_all = {}

    data_all = f_up.unpack_binary_data(
        file,
        data_params.daughterboard_number,
        data_params.motherboard_number,
        data_params.firmware_version,
        data_params.timestamps,
        data_params.include_offset,
        data_params.apply_calibration,
    )

    deltas_all = cd.calculate_differences_2212(data_all, pixels, pix_coor)

    data_for_plot_df = pd.DataFrame.from_dict(deltas_all, orient="index")

    result_queue.put(data_for_plot_df.T)


def _calculate_timestamps_differences_full_sensor(
    files: str,
    result_queue: multiprocessing.Queue,
    data_params: DataParamsConfig,
):
    # TODO add option for collecting from more than just two pixels
    # TODO use pixel handling function for modularity (if possible)
    """Calculate and save timestamp differences into '.feather' file.

    Unpacks data into a dictionary, calculates timestamp differences for
    the requested pixels and saves them into a '.feather' table. Works with
    firmware version 2212. Analyzes data from both sensor halves/both
    FPGAs, hence the two input parameters for LinoSPAD2 motherboards.

    Parameters
    ----------
    path : str
        Path to where two folders with data from both motherboards
        are. The folders should be named after the motherboards.
    pixels : list
        List of two pixels, one from each sensor half.
    rewrite : bool
        Switch for rewriting the '.feather' file if it already exists.
    daughterboard_number : str
        LinoSPAD2 daughterboard number.
    motherboard_number1 : str
        First LinoSPAD2 motherboard (FPGA) number.
    motherboard_number2 : str
        Second LinoSPAD2 motherboard (FPGA) number.
    firmware_version: str
        LinoSPAD2 firmware version. Versions "2212s" (skip) and "2212b"
        (block) are recognized.
    timestamps : int, optional
        Number of timestamps per acquisition cycle per pixel. The default
        is 512.
    delta_window : float, optional
        Size of a window to which timestamp differences are compared.
        Differences in that window are saved. The default is 50e3 (50 ns).
    app_mask : bool, optional
        Switch for applying the mask for hot pixels. The default is True.
    include_offset : bool, optional
        Switch for applying offset calibration. The default is True.
    apply_calibration : bool, optional
        Switch for applying TDC and offset calibration. If set to 'True'
        while include_offset is set to 'False', only the TDC calibration is
        applied. The default is True.
    absolute_timestamps : bool, optional
        Switch for unpacking data that were collected together with
        absolute timestamps. The default is False.

    Raises
    ------
    TypeError
        Only boolean values of 'rewrite' and string values of
        'daughterboard_number', 'motherboard_number', and 'firmware_version'
        are accepted. The first error is raised so that the plot does not
        accidentally get rewritten in the case no clear input was given.
    FileNotFoundError
        Raised if data from the first LinoSPAD2 motherboard were not
        found.
    FileNotFoundError
        Raised if data from the second LinoSPAD2 motherboard were not
        found.
    """
    # # parameter type check
    # if isinstance(data_params.pixels, list) is False:
    #     raise TypeError(
    #         "'pixels' should be a list of integers or a list of two lists"
    #     )
    # if isinstance(data_params.firmware_version, str) is False:
    #     raise TypeError(
    #         "'firmware_version' should be string, '2212s', '2212b' or" "'2208'"
    #     )
    # if isinstance(data_params.rewrite, bool) is False:
    #     raise TypeError("'rewrite' should be boolean")
    # if isinstance(data_params.daughterboard_number, str) is False:
    #     raise TypeError("'daughterboard_number' should be string")

    # Define matrix of pixel coordinates, where rows are numbers of TDCs
    # and columns are the pixels that connected to these TDCs
    if data_params.firmware_version == "2212s":
        pix_coor = np.arange(256).reshape(4, 64).T
    elif data_params.firmware_version == "2212b":
        pix_coor = np.arange(256).reshape(64, 4)
    else:
        print("\nFirmware version is not recognized.")
        sys.exit()

    # abs_tmsp_list1 = []
    # abs_tmsp_list2 = []

    deltas_all = {}

    # First board, unpack data
    os.chdir(
        os.path.join(data_params.path, f"{data_params.motherboard_number1}")
    )

    if not data_params.absolute_timestamps:
        data_all1 = f_up.unpack_binary_data(
            files[0],
            data_params.daughterboard_number,
            data_params.motherboard_number1,
            data_params.firmware_version,
            data_params.timestamps,
            data_params.include_offset,
            data_params.apply_calibration,
        )
    else:
        (
            data_all1,
            abs_tmsp1,
        ) = f_up.unpack_binary_data_with_absolute_timestamps(
            files[0],
            data_params.daughterboard_number,
            data_params.motherboard_number1,
            data_params.firmware_version,
            data_params.timestamps,
            data_params.include_offset,
            data_params.apply_calibration,
        )
    # abs_tmsp_list1.append(abs_tmsp1)

    # Collect indices of cycle ends (the '-2's)
    cycle_ends1 = np.where(data_all1[0].T[1] == -2)[0]
    cyc1 = np.argmin(
        np.abs(cycle_ends1 - np.where(data_all1[:].T[1] > 0)[0].min())
    )
    if cycle_ends1[cyc1] > np.where(data_all1[:].T[1] > 0)[0].min():
        cycle_start1 = cycle_ends1[cyc1 - 1]
    else:
        cycle_start1 = cycle_ends1[cyc1]

    # Second board, unpack data
    os.chdir(
        os.path.join(data_params.path, f"{data_params.motherboard_number2}")
    )

    if not data_params.absolute_timestamps:
        data_all2 = f_up.unpack_binary_data(
            files[1],
            data_params.daughterboard_number,
            data_params.motherboard_number2,
            data_params.firmware_version,
            data_params.timestamps,
            data_params.include_offset,
            data_params.apply_calibration,
        )
    else:
        (
            data_all2,
            abs_tmsp2,
        ) = f_up.unpack_binary_data_with_absolute_timestamps(
            files[1],
            data_params.daughterboard_number,
            data_params.motherboard_number2,
            data_params.firmware_version,
            data_params.timestamps,
            data_params.include_offset,
            data_params.apply_calibration,
        )
    # abs_tmsp_list2.append(abs_tmsp2)

    # Collect indices of cycle ends (the '-2's)
    cycle_ends2 = np.where(data_all2[0].T[1] == -2)[0]
    cyc2 = np.argmin(
        np.abs(cycle_ends2 - np.where(data_all2[:].T[1] > 0)[0].min())
    )
    if cycle_ends2[cyc2] > np.where(data_all2[:].T[1] > 0)[0].min():
        cycle_start2 = cycle_ends2[cyc2 - 1]
    else:
        cycle_start2 = cycle_ends2[cyc2]

    pix_left_peak = data_params.pixels[0]
    # The following piece of code take any value for the pixel from
    # the second motherboard, either in terms of full sensor (so
    # a value >=256) or in terms of single sensor half

    if data_params.pixels[1] >= 256:
        if data_params.pixels[1] > 256 + 127:
            pix_right_peak = 255 - (data_params.pixels[1] - 256)
        else:
            pix_right_peak = data_params.pixels[1] - 256 + 128
    elif data_params.pixels[1] > 127:
        pix_right_peak = 255 - data_params.pixels[1]
    else:
        pix_right_peak = data_params.pixels[1] + 128

    # Get the data from the requested pixel only
    deltas_all[f"{pix_left_peak},{pix_right_peak}"] = []
    tdc1, pix_c1 = np.argwhere(pix_coor == pix_left_peak)[0]
    pix1 = np.where(data_all1[tdc1].T[0] == pix_c1)[0]
    tdc2, pix_c2 = np.argwhere(pix_coor == pix_right_peak)[0]
    pix2 = np.where(data_all2[tdc2].T[0] == pix_c2)[0]

    # Data from one of the board should be shifted as data collection
    # on one of the board is started later
    if cycle_start1 > cycle_start2:
        cyc = len(data_all1[0].T[1]) - cycle_start1 + cycle_start2
        cycle_ends1 = cycle_ends1[cycle_ends1 >= cycle_start1]
        cycle_ends2 = np.intersect1d(
            cycle_ends2[cycle_ends2 >= cycle_start2],
            cycle_ends2[cycle_ends2 <= cyc],
        )
    else:
        cyc = len(data_all1[0].T[1]) - cycle_start2 + cycle_start1
        cycle_ends2 = cycle_ends2[cycle_ends2 >= cycle_start2]
        cycle_ends1 = np.intersect1d(
            cycle_ends1[cycle_ends1 >= cycle_start1],
            cycle_ends1[cycle_ends1 <= cyc],
        )
    # Get timestamps for both pixels in the given cycle
    for cyc in range(len(cycle_ends1) - 1):
        pix1_ = pix1[
            np.logical_and(
                pix1 >= cycle_ends1[cyc], pix1 < cycle_ends1[cyc + 1]
            )
        ]
        if not np.any(pix1_):
            continue
        pix2_ = pix2[
            np.logical_and(
                pix2 >= cycle_ends2[cyc], pix2 < cycle_ends2[cyc + 1]
            )
        ]

        if not np.any(pix2_):
            continue
        # Calculate delta t
        tmsp1 = data_all1[tdc1].T[1][
            pix1_[np.where(data_all1[tdc1].T[1][pix1_] > 0)[0]]
        ]
        tmsp2 = data_all2[tdc2].T[1][
            pix2_[np.where(data_all2[tdc2].T[1][pix2_] > 0)[0]]
        ]
        for t1 in tmsp1:
            deltas = tmsp2 - t1
            ind = np.where(np.abs(deltas) < data_params.delta_window)[0]
            deltas_all[f"{pix_left_peak},{pix_right_peak}"].extend(deltas[ind])

    # Save data to a feather file in a cycle so data is not lost
    # in the case of failure close to the end
    data_for_plot_df = pd.DataFrame.from_dict(deltas_all, orient="index")
    del deltas_all

    result_queue.put(data_for_plot_df.T)


def _write_results_to_feather(result_queue, feather_file, lock) -> None:
    """Save or append data to a Feather file.

    Parameters
    ----------
    result_queue : multiprocessing.Queue
        Interprocess queue containing data to be saved or appended to
        the Feather file.
    feather_file : str
        Absolute path to the Feather file.
    lock : multiprocessing.Lock
        Shared interprocess lock for synchronizing access to the Feather
        file.

    Notes
    -----
    The data in `result_queue` is retrieved and combined with the
    existing data in the Feather file, or a new file is created if it
    doesn't exist.
    The process is synchronized using the shared `lock` to prevent
    conflicts during file access.
    """
    while True:
        result_df = result_queue.get()
        if result_df is None:
            break

        # Use a lock to prevent conflicts when writing to the file
        with lock:
            if os.path.exists(feather_file):
                existing_data = ft.read_feather(feather_file)
                combined_data = pd.concat([existing_data, result_df], axis=0)
            else:
                combined_data = result_df.copy()

            # Reset the index to avoid issues during feather.write_feather
            combined_data.reset_index(drop=True, inplace=True)

            # Write the combined data to the Feather file
            ft.write_feather(combined_data, feather_file)


def _write_results_to_txt(result_queue_txt, txt_file, lock) -> None:
    """Save or append data to a txt file.

    Parameters
    ----------
    result_queue : multiprocessing.Queue
        Interprocess queue containing data to be saved or appended to
        the txt file.
    txt_file : str
        Absolute path to the txt file.
    lock : multiprocessing.Lock
        Shared interprocess lock for synchronizing access to the Feather
        file.

    Notes
    -----
    The data in `result_queue` is retrieved and combined with the
    existing data in the txt file, or a new file is created if it
    doesn't exist.
    The process is synchronized using the shared `lock` to prevent
    conflicts during file access.
    """
    while True:
        result_df = result_queue_txt.get()
        if result_df is None:
            break

        # Use a lock to prevent conflicts when writing to the file
        with lock:
            if os.path.exists(txt_file):
                existing_data = np.genfromtxt(
                    txt_file, delimiter="\t", dtype=int
                )
                combined_data = existing_data + result_df
            else:
                combined_data = result_df.copy()

            # Write the combined data to the Feather file
            np.savetxt(
                txt_file,
                combined_data,
                fmt="%d",
                delimiter="\t",
            )


def calculate_and_save_timestamp_differences_mp(
    path: str,
    pixels: list,
    rewrite: bool,
    daughterboard_number: str,
    motherboard_number: str,
    firmware_version: str,
    timestamps: int,
    delta_window: float = 50e3,
    app_mask: bool = True,
    include_offset: bool = False,
    apply_calibration: bool = True,
    chunksize: int = 20,
    number_of_cores: int = 4,
    maxtasksperchild: int = 1000,
) -> None:
    """Unpack data and collect timestamps differences using all CPU cores.

    Unpack data files and collect timestamps differences using all
    available CPU cores to speed up the process, while saving the results
    to a single Feather file on the go.

    Parameters
    ----------
    path : str
        Path to data files.
    pixels : list
        List of pixel numbers for which the timestamp differences should
        be calculated and saved or list of two lists with pixel numbers
        for peak vs. peak calculations.
    rewrite : bool
        Switch for rewriting the '.feather' file if it already exists.
    daughterboard_number : str
        LinoSPAD2 daughterboard number.
    motherboard_number : str
        LinoSPAD2 motherboard (FPGA) number, including the "#".
    firmware_version: str
        LinoSPAD2 firmware version. Versions "2212s" (skip) and "2212b"
        (block) are recognized.
    timestamps : int, optional
        Number of timestamps per acquisition cycle per pixel.
    delta_window : float, optional
        Size of a window to which timestamp differences are compared.
        Differences in that window are saved. The default is 50e3 (50 ns).
    app_mask : bool, optional
        Switch for applying the mask for hot pixels. The default is True.
    include_offset : bool, optional
        Switch for applying offset calibration. The default is True.
    apply_calibration : bool, optional
        Switch for applying TDC and offset calibration. If set to 'True'
        while apply_offset_calibration is set to 'False', only the TDC
        calibration is applied. The default is True.
    chunksize : int, optional
        Number of files processed in each iteration. The default is 20.
    number_of_cores : int, optional
        Number of cores to use for multiprocessing. The default is 4.
    maxtasksperchild : int, optional
        Number of tasks per core. The default is 1000.
    Raises
    ------
    TypeError
        Only boolean values of 'rewrite' and string values of
        'daughterboard_number', 'motherboard_number', and 'firmware_version'
        are accepted. The first error is raised so that the plot does not
        accidentally get rewritten in the case no clear input was given.

    Returns
    -------
    None.
    """
    # parameter type check
    if isinstance(pixels, list) is False:
        raise TypeError(
            "'pixels' should be a list of integers or a list of two lists"
        )
    if isinstance(firmware_version, str) is False:
        raise TypeError(
            "'firmware_version' should be string, '2212s', '2212b' or '2208'"
        )
    if isinstance(rewrite, bool) is False:
        raise TypeError("'rewrite' should be boolean")
    if isinstance(daughterboard_number, str) is False:
        raise TypeError("'daughterboard_number' should be string")

    # Generate a dataclass object
    data_params = DataParamsConfig(
        pixels=pixels,
        daughterboard_number=daughterboard_number,
        motherboard_number=motherboard_number,
        firmware_version=firmware_version,
        timestamps=timestamps,
        delta_window=delta_window,
        app_mask=app_mask,
        include_offset=include_offset,
        apply_calibration=apply_calibration,
        absolute_timestamps=True,
    )

    os.chdir(path)

    # Find all LinoSPAD2 data files
    # files = sorted(glob.glob("*.dat"))
    files = glob.glob("*.dat*")
    files.sort(key=os.path.getmtime)
    # Get the resulting Feather file name based on the data files
    # found
    feather_file_name = files[0][:-4] + "-" + files[-1][:-4] + ".feather"
    # Construct absolute path to the Feather file
    feather_file = os.path.join(path, "delta_ts_data", feather_file_name)
    # Handle the rewrite parameter based on the file existence to avoid
    # accidental file overwritting
    utils.file_rewrite_handling(feather_file, rewrite)

    with multiprocessing.Manager() as manager:
        shared_result_queue = manager.Queue()
        shared_lock = manager.Lock()

        with multiprocessing.Pool(
            processes=number_of_cores, maxtasksperchild=maxtasksperchild
        ) as pool:
            # Start the writer process
            writer_process = multiprocessing.Process(
                target=_write_results_to_feather,
                args=(shared_result_queue, feather_file, shared_lock),
            )
            writer_process.start()

            # Create a partial function with fixed arguments for
            # process_file
            partial_process_file = functools.partial(
                _calculate_timestamps_differences,
                result_queue=shared_result_queue,
                data_params=data_params,
            )

            # Start the multicore analysis of the files
            pool.map(partial_process_file, files, chunksize=chunksize)

            # pool.close()
            # pool.join()
            # Use tqdm to create a progress bar for the file processing
            # TODO Can't configure tqdm to update the progress bar
            # correctly while using chunksize in pool

            # with tqdm(total=len(files), desc="Processing files") as pbar:
            #     for _ in pool.imap_unordered(partial_process_file, files):
            #         pbar.update(1)

            # Signal the writer process that no more results will be
            # added to the queue
            shared_result_queue.put(None)
            writer_process.join()


def calculate_and_save_timestamp_differences_full_sensor_mp(
    path: str,
    pixels: list,
    rewrite: bool,
    daughterboard_number: str,
    motherboard_number1: str,
    motherboard_number2: str,
    firmware_version: str,
    timestamps: int,
    delta_window: float = 50e3,
    app_mask: bool = True,
    include_offset: bool = False,
    apply_calibration: bool = True,
    absolute_timestamps: bool = True,
    chunksize: int = 20,
) -> None:
    """Unpack data and collect timestamps differences using all CPU cores.

    Unpack data files and collect timestamps differences using all
    available CPU cores to speed up the process, while saving the results
    to a single Feather file on the go. Version for analyzing data
    from both halves of the sensor.

    Parameters
    ----------
    path : str
        Path to data files.
    pixels : list
        List of pixel numbers for which the timestamp differences should
        be calculated and saved or list of two lists with pixel numbers
        for peak vs. peak calculations.
    rewrite : bool
        Switch for rewriting the '.feather' file if it already exists.
    daughterboard_number : str
        LinoSPAD2 daughterboard number.
    motherboard_number1 : str
        First LinoSPAD2 motherboard (FPGA) number.
    motherboard_number2 : str
        Second LinoSPAD2 motherboard (FPGA) number.
    firmware_version: str
        LinoSPAD2 firmware version. Versions "2212s" (skip) and "2212b"
        (block) are recognized.
    timestamps : int, optional
        Number of timestamps per acquisition cycle per pixel.
    delta_window : float, optional
        Size of a window to which timestamp differences are compared.
        Differences in that window are saved. The default is 50e3 (50 ns).
    app_mask : bool, optional
        Switch for applying the mask for hot pixels. The default is True.
    include_offset : bool, optional
        Switch for applying offset calibration. The default is True.
    apply_calibration : bool, optional
        Switch for applying TDC and offset calibration. If set to 'True'
        while apply_offset_calibration is set to 'False', only the TDC
        calibration is applied. The default is True.
    absolute_timestamps : bool, False
        Indicator for data with absolute timestamps. The default is
        False.
    chunksize : int, optional
        Number of files processed in each iteration. The default is 20.

    Raises
    ------
    TypeError
        Only boolean values of 'rewrite' and string values of
        'daughterboard_number', 'motherboard_number', and 'firmware_version'
        are accepted. The first error is raised so that the plot does not
        accidentally get rewritten in the case no clear input was given.

    Returns
    -------
    None.
    """
    # parameter type check
    if isinstance(pixels, list) is False:
        raise TypeError(
            "'pixels' should be a list of integers or a list of two lists"
        )
    if isinstance(firmware_version, str) is False:
        raise TypeError(
            "'firmware_version' should be string, '2212s', '2212b' or '2208'"
        )
    if isinstance(rewrite, bool) is False:
        raise TypeError("'rewrite' should be boolean")
    if isinstance(daughterboard_number, str) is False:
        raise TypeError("'daughterboard_number' should be string")

    # Generate a dataclass object
    data_params = DataParamsConfig(
        pixels=pixels,
        path=path,
        daughterboard_number=daughterboard_number,
        motherboard_number1=motherboard_number1,
        motherboard_number2=motherboard_number2,
        firmware_version=firmware_version,
        timestamps=timestamps,
        delta_window=delta_window,
        app_mask=app_mask,
        include_offset=include_offset,
        apply_calibration=apply_calibration,
        absolute_timestamps=absolute_timestamps,
    )

    # Check the data from the first FPGA board
    try:
        os.chdir(os.path.join(path, f"{motherboard_number1}"))
    except FileNotFoundError as exc:
        raise FileNotFoundError(
            f"Data from {motherboard_number1} not found"
        ) from exc
    # files_all1 = sorted(glob.glob("*.dat*"))
    files_all1 = glob.glob("*.dat*")
    files_all1.sort(key=os.path.getmtime)
    out_file_name = files_all1[0][:-4]

    # Check the data from the second FPGA board
    try:
        # os.chdir(f"{motherboard_number2}")
        os.chdir(os.path.join(path, f"{motherboard_number2}"))
    except FileNotFoundError as exc:
        raise FileNotFoundError(
            f"Data from {motherboard_number2} not found"
        ) from exc
    # files_all2 = sorted(glob.glob("*.dat*"))
    files_all2 = glob.glob("*.dat*")
    files_all2.sort(key=os.path.getmtime)
    out_file_name = out_file_name + "-" + files_all2[-1][:-4]

    # Check if '.feather' file with timestamps differences already
    # exists
    out_file_name = os.path.join(path, "delta_ts_data", out_file_name)
    feather_file = f"{out_file_name}.feather"

    utils.file_rewrite_handling(feather_file, rewrite)

    files = zip(files_all1, files_all2)

    with multiprocessing.Manager() as manager:
        shared_result_queue = manager.Queue()
        shared_lock = manager.Lock()

        with multiprocessing.Pool() as pool:
            # Start the writer process
            writer_process = multiprocessing.Process(
                target=_write_results_to_feather,
                args=(shared_result_queue, feather_file, shared_lock),
            )
            writer_process.start()

            # Create a partial function with fixed arguments for
            # process_file
            partial_process_file = functools.partial(
                _calculate_timestamps_differences_full_sensor,
                result_queue=shared_result_queue,
                data_params=data_params,
            )

            # Start the multicore analysis of the files
            pool.map(partial_process_file, list(files), chunksize=chunksize)

            # Use tqdm to create a progress bar for the file processing
            # TODO Can't configure tqdm to update the progress bar
            # correctly while using chunksize in pool

            # with tqdm(total=len(files), desc="Processing files") as pbar:
            #     for _ in pool.imap_unordered(partial_process_file, files):
            #         pbar.update(1)

            # Signal the writer process that no more results will be
            # added to the queue
            shared_result_queue.put(None)
            writer_process.join()


def _compact_share_collect_data(
    file: str,
    result_queue_feather: multiprocessing.Queue,
    result_queue_txt: multiprocessing.Queue,
    data_params: DataParamsConfig,
):
    """Collect delta timestamp differences and sensor population from
    unpacked data, saving results to '.feather' and '.txt' files.

    This function processes data from a single .dat file, calculates
    timestamp differences and collects sensor population numbers. The
    resulting arrays are put each to a separate multiprocessing queue
    for saving via an appropriate function.

    Parameters
    ----------
    file : str
        Path to the data file to be processed.
    result_queue_feather : multiprocessing.Queue
        Queue for storing the timestamp differences in '.feather' format.
    result_queue_txt : multiprocessing.Queue
        Queue for storing the sensor population in '.txt' format.
    data_params : DataParamsConfig
        Configuration object containing parameters for data processing.

    Returns
    -------
    None.
    """

    os.chdir(data_params.path)

    # Collect the data for the required pixels

    # for transforming pixel number into TDC number + pixel
    # coordinates in that TDC
    if data_params.firmware_version == "2212s":
        pix_coor = np.arange(256).reshape(4, 64).T
    elif data_params.firmware_version == "2212b":
        pix_coor = np.arange(256).reshape(64, 4)
    else:
        sys.exit()

    # Prepare array for sensor population
    valid_per_pixel = np.zeros(256, dtype=int)

    # Prepare a dictionary for timestamp differences
    deltas_all = {}

    # Unpack data
    if not data_params.absolute_timestamps:
        data_all = f_up.unpack_binary_data(
            file,
            data_params.daughterboard_number,
            data_params.motherboard_number,
            data_params.firmware_version,
            data_params.timestamps,
            data_params.include_offset,
            data_params.apply_calibration,
        )
    else:
        data_all, _ = f_up.unpack_binary_data_with_absolute_timestamps(
            file,
            data_params.daughterboard_number,
            data_params.motherboard_number,
            data_params.firmware_version,
            data_params.timestamps,
            data_params.include_offset,
            data_params.apply_calibration,
        )

    deltas_all = cd.calculate_differences_2212(
        data_all, data_params.pixels, pix_coor, data_params.delta_window
    )

    # Collect sensor population
    for k in range(256):
        tdc, pix = np.argwhere(pix_coor == k)[0]
        valid_per_pixel[k] += np.count_nonzero(data_all[tdc][:, 0] == pix)

    # Save data as a .feather file in a cycle so data is not lost
    # in the case of failure close to the end
    data_for_plot_df = pd.DataFrame.from_dict(deltas_all, orient="index")
    del deltas_all

    result_queue_feather.put(data_for_plot_df.T)
    result_queue_txt.put(valid_per_pixel)


def compact_share_mp(
    path: str,
    pixels: list,
    rewrite: bool,
    daughterboard_number: str,
    motherboard_number: str,
    firmware_version: str,
    timestamps: int,
    delta_window: float = 50e3,
    include_offset: bool = False,
    apply_calibration: bool = True,
    absolute_timestamps: bool = False,
    chunksize=None,
):
    """Collect timestamp differences and sensor population using
    multiprocessing, saving results to '.feather' and '.txt' files.

    This function parallelizes the processing of data in the specified
    path using multiprocessing. It calculates timestamp differences and
    sensor population for the specified pixels based on the provided
    parameters, saving the timestamp differences to a '.feather' file and
    the sensor population to a '.txt' file. Both files are then zipped for
    compact output ready to share.

    Parameters
    ----------
    path : str
        Path to the directory containing data files.
    pixels : list
        List of pixel numbers for which the timestamp differences should
        be calculated and saved, or list of two lists with pixel numbers
        for peak vs. peak calculations.
    rewrite : bool
        Switch for rewriting the '.feather' file if it already exists.
    daughterboard_number : str
        LinoSPAD2 daughterboard number.
    motherboard_number : str
        LinoSPAD2 motherboard (FPGA) number, including the "#".
    firmware_version: str
        LinoSPAD2 firmware version. Accepted values are "2212s" (skip)
        and "2212b" (block).
    timestamps : int
        Number of timestamps per acquisition cycle per pixel.
    delta_window : float, optional
        Size of a window (in nanoseconds) to which timestamp differences
        are compared (default is 50e3 nanoseconds).
    include_offset : bool, optional
        Switch for applying offset calibration (default is True).
    apply_calibration : bool, optional
        Switch for applying TDC and offset calibration. If set to 'True'
        while include_offset is set to 'False', only the TDC calibration is
        applied (default is True).
    absolute_timestamps : bool, optional
        Indicator for data with absolute timestamps (default is False).
    chunksize : int, optional
        The number of data points processed in each batch by each worker.
        If None, the default chunk size is determined automatically.

    Raises
    ------
    TypeError
        Only boolean values of 'rewrite' and string values of
        'firmware_version' are accepted. The first error is raised so
        that the files are not accidentally overwritten in the case of
        unclear input.

    Returns
    -------
    None.
    """

    # parameter type check
    if isinstance(pixels, list) is False:
        raise TypeError(
            "'pixels' should be a list of integers or a list of two lists"
        )
    if isinstance(firmware_version, str) is False:
        raise TypeError(
            "'firmware_version' should be string, '2212s', '2212b' or '2208'"
        )
    if isinstance(rewrite, bool) is False:
        raise TypeError("'rewrite' should be boolean")
    if isinstance(daughterboard_number, str) is False:
        raise TypeError("'daughterboard_number' should be string")
    if isinstance(motherboard_number, str) is False:
        raise TypeError("'motherboard_number' should be string")

    # Generate a dataclass object
    data_params = DataParamsConfig(
        pixels=pixels,
        path=path,
        daughterboard_number=daughterboard_number,
        motherboard_number=motherboard_number,
        firmware_version=firmware_version,
        timestamps=timestamps,
        delta_window=delta_window,
        include_offset=include_offset,
        apply_calibration=apply_calibration,
        absolute_timestamps=absolute_timestamps,
    )

    os.chdir(path)

    # Find all LinoSPAD2 data files
    # files = sorted(glob.glob("*.dat"))
    files = glob.glob("*.dat*")
    files.sort(key=os.path.getmtime)
    # Get the resulting Feather file name based on the data files
    # found
    feather_file_name = files[0][:-4] + "-" + files[-1][:-4] + ".feather"
    txt_file_name = files[0][:-4] + "-" + files[-1][:-4] + ".txt"
    # Construct absolute path to the Feather file
    feather_file = os.path.join(path, "compact_share", feather_file_name)
    txt_file = os.path.join(path, "compact_share", txt_file_name)
    # Handle the rewrite parameter based on the file existence to avoid
    # accidental file overwritting
    utils.file_rewrite_handling(feather_file, rewrite)
    utils.file_rewrite_handling(txt_file, rewrite)

    with multiprocessing.Manager() as manager:
        shared_result_feather = manager.Queue()
        shared_result_txt = manager.Queue()
        shared_lock_feather = manager.Lock()
        shared_lock_txt = manager.Lock()

        with multiprocessing.Pool() as pool:
            # Start the writer process
            writer_process_feather = multiprocessing.Process(
                target=_write_results_to_feather,
                args=(
                    shared_result_feather,
                    feather_file,
                    shared_lock_feather,
                ),
            )

            writer_process_txt = multiprocessing.Process(
                target=_write_results_to_txt,
                args=(shared_result_txt, txt_file, shared_lock_txt),
            )

            writer_process_feather.start()
            writer_process_txt.start()

            # Create a partial function with fixed arguments for
            # process_file
            partial_process_file = functools.partial(
                _compact_share_collect_data,
                result_queue_feather=shared_result_feather,
                result_queue_txt=shared_result_txt,
                data_params=data_params,
            )

            # Start the multicore analysis of the files
            if chunksize is None:
                pool.map(partial_process_file, files)

            else:
                pool.map(partial_process_file, files, chunksize=chunksize)

            # Signal the writer process that no more results will be
            # added to the queue
            shared_result_feather.put(None)
            shared_result_txt.put(None)
            writer_process_feather.join()
            writer_process_txt.join()

        # Create a ZipFile Object
        os.chdir(os.path.join(path, "compact_share"))
        out_file_name, _ = os.path.splitext(feather_file_name)

        with ZipFile(f"{out_file_name}.zip", "w") as zip_object:
            # Adding files that need to be zipped
            zip_object.write(f"{out_file_name}.feather")
            zip_object.write(f"{out_file_name}.txt")

            print(
                "\n> > > Timestamp differences are saved as {feather_file} and "
                "sensor population as {txt_file} in "
                "{path} < < <".format(
                    feather_file=feather_file,
                    txt_file=txt_file,
                    path=path + "\delta_ts_data",
                )
            )



import os
import shutil
import unittest

from LinoSPAD2.functions.sensor_plot import (
    plot_sensor_population,
    plot_single_pix_hist,
)


class TestPlotScripts(unittest.TestCase):
    def setUp(self):
        self.path = "tests/test_data"
        self.pix = 15
        self.daughterboard_number = "NL11"
        self.motherboard_number = "#33"
        self.firmware_version = "2212b"
        self.timestamps = 300

    def test_a_plot_pixel_hist(self):
        # Positive test case
        os.chdir(
            r"{}".format(os.path.dirname(os.path.realpath(__file__)) + "/..")
        )

        plot_single_pix_hist(
            self.path,
            self.pix,
            self.daughterboard_number,
            self.motherboard_number,
            self.firmware_version,
            self.timestamps,
        )
        self.assertTrue(
            os.path.exists(
                "results/single pixel histograms/test_data_2212b.dat, pixel 15.png"
            )
        )

    def test_b_plot_sen_pop(self):
        # Positive test case
        os.chdir(
            r"{}".format(os.path.dirname(os.path.realpath(__file__)) + "/..")
        )
        plot_sensor_population(
            self.path,
            self.daughterboard_number,
            self.motherboard_number,
            self.firmware_version,
            self.timestamps,
            scale="linear",
            style="-o",
            show_fig=True,
            app_mask=True,
        )
        self.assertTrue(
            os.path.isfile(
                "results/sensor_population/test_data_2212b-test_data_2212b.png"
            )
        )

    # TODO: data for SPDC with background needed
    # def test_c_plot_spdc(self):
    #     # Positive test case
    #     os.chdir(r"{}".format(os.path.realpath(__file__) + "/../.."))
    #     plot_spdc(self.path, self.board_number, self.timestamps, show_fig=True)
    #     self.assertTrue(
    #         os.path.exists(
    #             "results/test_data_2212b-test_data_2212b_SPDC_counts.png"
    #         )
    #     )

    #     # Negative test case
    #     with self.assertRaises(TypeError):
    #         plot_spdc(
    #             self.path, self.board_number, self.timestamps, show_fig="True"
    #         )

    def tearDownClass():
        # Clean up after tests
        os.chdir(r"{}".format(os.path.dirname(os.path.realpath(__file__))))
        shutil.rmtree("test_data/results")


if __name__ == "__main__":
    unittest.main()



"""Module for analyzing cross-talk of LinoSPAD2.

A set of functions to calculate and save, collect and plot the
cross-talk data for the given data sets.

This file can also be imported as a module and contains the following
functions:

    * collect_dcr_by_file - calculate the DCR for each file in the
    given folder and save the result into a .pkl file. Calculate
    average and median DCR with and without masking the hot pixels.

    * plot_dcr_histogram_and_stability - plot the DCR vs file to check
    setup stability, and plot histogram of DCR averaged over files
    together with integral over pixels

    * zero_to_cross_talk_collect - collect timestamp differences from
    the '.dat' with the cross-talk data (preferably, noise only).
    Timestamp differences are collected for the given hot pixels plus
    20 pixels to each side of the hot pixel where it is possible. The
    result is saved into a '.feather' file.

    * zero_to_cross_talk_plot - collect the timestamp differences from
    the '.feather' file and plot the cross-talk peaks applying a 
    Gaussian fit where possible, plot the average cross-talk probability
    vs distance from the hot pixel for each pair of aggressor-victim
    pixels, and finally plot the average cross-talk probability
    for the whole sensor half.

"""

import glob
import os
import pickle
import sys
from typing import List

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from pyarrow import feather as ft
from scipy.optimize import curve_fit
from tqdm import tqdm

from LinoSPAD2.functions import calc_diff as cd
from LinoSPAD2.functions import sensor_plot
from LinoSPAD2.functions import unpack as f_up
from LinoSPAD2.functions import utils


def _collect_cross_talk(
    path: str,
    pixels: List[int],
    rewrite: bool,
    daughterboard_number: str,
    motherboard_number: str,
    firmware_version: str,
    timestamps: int = 1000,
    delta_window: float = 50e3,
    include_offset: bool = False,
    apply_calibration: bool = True,
    absolute_timestamps: bool = False,
    correct_pix_address: bool = False,
):
    """Collect timestamp differences from cross-talk data.

    Collect timestamp differences from cross-talk data for the given
    list of pixels.

    Parameters
    ----------
    path : str
        Path to the folder with '.dat' data files.
    pixels : List[int]
        List of pixels to collect timestamp differences for, where the
        first pixel is the aggressor.
    rewrite : bool
        Switch for rewriting the '.feather' file with timestamp
        differences. Used for conscious start of an hours-long data
        processing and avoiding rewriting/deleting already existing
        files.
    daughterboard_number : str
        LinoSPAD2 daughterboard number.
    motherboard_number : str
        LinoSPAD2 motherboard (FPGA) number, including the '#'.
    firmware_version : str
        LinoSPAD2 firmware version.
    timestamps : int, optional
        Number of timestamps per TDC per cycle used during data
        collection. The default is 1000.
    include_offset : bool
        Switch for including the offset calibration.
    delta_window : float, optional
        Window size for collecting timestamp differences. The default
        is 50e3.
    apply_calibration : bool, optional
        If True, apply calibration to the collected data. The default
        is True.
    absolute_timestamps : bool, optional
        Indicator of data collected with absolute timestamps. The
        default is False.
    correct_pix_address : bool, optional
        Correct pixel address for the FPGA board on side 23. Here
        used to reverse the correction. The default is False.

    Raises
    ------
    TypeError
        Raised if any of the following conditions are met:
        - `pixels` is not a list of integers,
        - `firmware_version` is not recognized,
        - `rewrite` is not boolean,
        - `daughterboard_number` is not a string.
    """

    # parameter type check
    if isinstance(pixels, list) is False:
        raise TypeError(
            "'pixels' should be a list of integers or a list of two lists"
        )
    if isinstance(firmware_version, str) is False:
        raise TypeError(
            "'firmware_version' should be string, '2212s', '2212b' or '2208'"
        )
    if isinstance(rewrite, bool) is False:
        raise TypeError("'rewrite' should be boolean")
    if isinstance(daughterboard_number, str) is False:
        raise TypeError("'daughterboard_number' should be string")
    os.chdir(path)

    # If requested, get the correct pixel address - must be used
    # for the motherboard on side '23'
    if correct_pix_address:
        for i, pixel in enumerate(pixels):
            if pixel > 127:
                pixels[i] = 255 - pixels[i]
            else:
                pixels[i] = pixels[i] + 128

    # files_all = sorted(glob.glob("*.dat*"))
    files_all = glob.glob("*.dat*")
    files_all.sort(key=os.path.getmtime)

    out_file_name = files_all[0][:-4] + "-" + files_all[-1][:-4]

    feather_file = os.path.join(
        path,
        "cross_talk_data",
        f"{out_file_name}_pixels_{pixels[0]}-{pixels[-1]}.feather",
    )

    utils.file_rewrite_handling(feather_file, rewrite)

    # Define matrix of pixel coordinates, where rows are numbers of TDCs
    # and columns are the pixels that connected to these TDCs
    if firmware_version == "2212s":
        pix_coor = np.arange(256).reshape(4, 64).T
    elif firmware_version == "2212b":
        print(
            "\nFor firmware version '2212b' cross-talk numbers "
            "would be incorrect, try data collected with '2212s'"
        )
        sys.exit()
    else:
        print("\nFirmware version is not recognized.")
        sys.exit()

    pixels_formatted = [[pixels[0]], pixels[1:]]

    for i in tqdm(range(len(files_all)), desc="Collecting data"):
        file = files_all[i]

        # Prepare a dictionary for output
        deltas_all = {}

        # Unpack data for the requested pixels into dictionary
        if not absolute_timestamps:
            data_all = f_up.unpack_binary_data(
                file,
                daughterboard_number,
                motherboard_number,
                firmware_version,
                timestamps,
                include_offset,
                apply_calibration,
            )
        else:
            data_all, _ = f_up.unpack_binary_data_with_absolute_timestamps(
                file,
                daughterboard_number,
                motherboard_number,
                firmware_version,
                timestamps,
                include_offset,
                apply_calibration,
            )

        # Collect timestamp differences for the given pixels
        # deltas_all = cd.calculate_differences_2212(
        #     data_all, pixels_formatted, pix_coor, delta_window
        # )
        deltas_all = cd.calculate_differences_2212_fast(
            data_all, pixels_formatted, pix_coor, delta_window
        )

        # Save data as a .feather file in a cycle so data is not lost
        # in the case of failure close to the end
        data_for_plot_df = pd.DataFrame.from_dict(deltas_all, orient="index")
        del deltas_all
        data_for_plot_df = data_for_plot_df.T
        try:
            os.chdir("cross_talk_data")
        except FileNotFoundError:
            os.mkdir("cross_talk_data")
            os.chdir("cross_talk_data")

        # Check if feather file exists
        feather_file = (
            f"{out_file_name}_pixels_{pixels[0]}-{pixels[-1]}.feather"
        )
        if os.path.isfile(feather_file):
            # Load existing feather file
            existing_data = ft.read_feather(feather_file)

            # Append new data to the existing feather file
            combined_data = pd.concat(
                [existing_data, data_for_plot_df], axis=0
            )
            ft.write_feather(combined_data, feather_file)

        else:
            # Save as a new feather file
            ft.write_feather(data_for_plot_df, feather_file)
        os.chdir("..")

    # Check, if the file was created
    if (
        os.path.isfile(
            path + f"/cross_talk_data/{out_file_name}_pixels_"
            f"{pixels[0]}-{pixels[-1]}.feather"
        )
        is True
    ):
        print(
            "\n> > > Timestamp differences are saved as"
            f"{out_file_name}_pixels_{pixels[0]}-{pixels[-1]}.feather in "
            f"{os.path.join(path, 'cross_talk_data')} < < <"
        )

    else:
        print("File wasn't generated. Check input parameters.")


def _plot_cross_talk_peaks(
    path: str,
    pixels: List,
    multiplier: int,
    window: float = 50e3,
    senpop: list = None,
    pix_on_left: bool = False,
    feather_file_name: str = "",
    show_plots: bool = False,
):
    """Plot cross-talk peaks and calculate peak population.

    Plots a histogram of timestamp differences collected from the
    cross-talk data, fits using Gaussian to extract sigma and peak
    position. These two numbers are then used to calculate the peak
    population as numbers of data points in a +/- 2*sigma window
    around the peak position minus background in the same window.

    Parameters
    ----------
    path : str
        Path to the folder with the '.dat' data files.
    pixels : list
        List of pixel number to plot cross-talk peaks for. The first
        one is the aggressor pixel.
    multiplier : int
        Multiplier of the LinoSPAD2 average timing bin of 17.857 ps for
        histograms of the timestamp differences.
    window : float
        Window length for the histogram range. The default is 50e3.
    senpop : list
        List with numbers of timestamps per pixel.
    pix_on_left : bool, optional
        Indicator of pixels to the left of the aggressor. The default is False.
    feather_file_name : str, optional
        Naming pattern of the feather files that can be used when
        actual data files are not available. The pattern should
        include the number of the first and the last data files, e.g.,
        "0000010000-0000012000". The default is "".
    show_plots : bool, optional
        Switch for showing the plots at the end. The default is False.

    Returns
    -------
    dict, dict
        Two dictionaries with distances in pixels from the aggressor
        pixel, where the first dictionary contains cross-talk
        probabilities and the second - the corresponding errors.
    """

    os.chdir(path)
    if feather_file_name == "":
        files_all1 = glob.glob("*.dat")
        files_all1.sort(key=os.path.getmtime)
        ft_file_name = files_all1[0][:-4] + "-" + files_all1[-1][:-4]
    else:
        ft_file_name = feather_file_name

    os.chdir(os.path.join(path, "cross_talk_data"))

    ft_file = glob.glob(
        f"{ft_file_name}_*{pixels[0]}" f"-{pixels[-1]}*.feather"
    )[0]

    ct_output = {}
    ct_err_output = {}

    for _, pix in enumerate(pixels[1:]):
        if pix_on_left:

            data_pix = (
                ft.read_feather(ft_file, columns=[f"{pix},{pixels[0]}"])
                .dropna()
                .values
            )
        else:
            data_pix = (
                ft.read_feather(ft_file, columns=[f"{pixels[0]},{pix}"])
                .dropna()
                .values
            )

        data_cut = data_pix[(data_pix > -window / 2) & (data_pix < window / 2)]

        if not data_cut.any().any():
            continue
        counts, bin_edges = np.histogram(
            data_cut,
            bins=(
                np.arange(
                    np.min(data_cut), np.max(data_cut), multiplier * 17.857
                )
            ),
        )
        if len(counts) < 1:
            continue
        bin_centers = (bin_edges - multiplier * 17.857 / 2)[1:]

        try:
            params, _ = curve_fit(
                utils.gaussian,
                bin_centers,
                counts,
                p0=[
                    np.max(counts),
                    bin_centers[np.argmax(counts)],
                    100,
                    np.median(counts),
                ],
            )
        except (RuntimeError, TypeError):
            continue

        # Cross-talk probability estimation
        if senpop is not None:
            aggressor_pix_tmsps = senpop[pixels[0]]
            victim_pix_tmsps = senpop[pix]

            peak_population = data_cut[
                (data_cut > params[1] - params[2] * 2)
                & (data_cut < params[1] + params[2] * 2)
            ]
            bckg = data_cut[
                (data_cut > params[1] + 15e3 - params[2] * 2)
                & (data_cut < params[1] + 15e3 + params[2] * 2)
            ]
            peak_population_err = np.sqrt(len(peak_population))
            bckg_err = np.sqrt(len(bckg))

            CT = (
                (len(peak_population) - len(bckg))
                * 100
                / (aggressor_pix_tmsps + victim_pix_tmsps)
            )

            CT_err = (
                np.sqrt(peak_population_err**2 + bckg_err**2)
                * 100
                / (aggressor_pix_tmsps + victim_pix_tmsps)
            )

        plt.rcParams.update({"font.size": 27})
        fig = plt.figure(figsize=(16, 10))
        plt.step(bin_centers, counts, color="rebeccapurple", label="Data")
        plt.plot(
            bin_centers,
            utils.gaussian(bin_centers, *params),
            color="darkorange",
            label="Fit",
        )
        if senpop is not None:
            plt.title(
                f"Cross-talk peak, pixels {pixels[0]},{pix}\n"
                f"Cross-talk is {CT:.1e}" + "\u00B1" + f"{CT_err:.1e}" + "%"
            )
        else:
            plt.title(f"Cross-talk peak, pixels {pixels[0]},{pix}")

        plt.xlabel("\u0394t (ps)")
        plt.ylabel("# of coincidences (-)")
        plt.legend()
        try:
            os.chdir(os.path.join(path, "results/ct_fit"))
        except FileNotFoundError as _:
            os.makedirs(os.path.join(path, "results/ct_fit"))
            os.chdir(os.path.join(path, "results/ct_fit"))
        plt.savefig(f"CT_fit_pixels_{pixels[0]},{pix}.png")
        os.chdir(os.path.join(path, "cross_talk_data"))

        if show_plots:
            plt.show()
        else:
            plt.close(fig)

        ct_output[f"{pixels[0], pix}"] = CT
        ct_err_output[f"{pixels[0], pix}"] = CT_err

    return ct_output, ct_err_output


def _plot_cross_talk_grid(
    path,
    pixels,
    step,
    window: int = 50e3,
    pix_on_left: bool = False,
    feather_file_name: str = "",
    show_plots: bool = False,
):
    """Plot grid of 4x5 of cross-talk peaks.

    Parameters
    ----------
    path : str
        Path to data files.
    pixels : list
        List of pixels to plot the grid for. The first is the aggressor
        pixel.
    step : int
        Multiplier of the LinoSPAD2 average timing bin of 17.857 ps for
        the histograms.
    window : int, optional
        Window size for the histogram range. The default is 50e3.
    pix_on_left : bool, optional
        Indicator that pixels provided are to the left of the aggressor.
        The default is False.
    feather_file_name : str, optional
        Naming pattern of the feather files that can be used when
        actual data files are not available. The pattern should
        include the number of the first and the last data files, e.g.,
        "0000010000-0000012000". The default is "".
    show_plots : bool, optional
        Switch for showing the plots at the end. The default is False.
    """

    os.chdir(path)
    if feather_file_name == "":
        files_all1 = glob.glob("*.dat")
        files_all1.sort(key=os.path.getmtime)
        ft_file_name = files_all1[0][:-4] + "-" + files_all1[-1][:-4]
    else:
        ft_file_name = feather_file_name

    os.chdir(os.path.join(path, "cross_talk_data"))

    ft_file = glob.glob(f"{ft_file_name}_*{pixels[0]}-{pixels[-1]}*.feather")[
        0
    ]

    fig, axes = plt.subplots(4, 5, figsize=(16, 10))
    plt.rcParams.update({"font.size": 27})

    for i, pix in enumerate(pixels[1:]):
        if pix_on_left:
            data_pix = (
                ft.read_feather(ft_file, columns=[f"{pix},{pixels[0]}"])
                .dropna()
                .values
            )
        else:
            data_pix = (
                ft.read_feather(ft_file, columns=[f"{pixels[0]},{pix}"])
                .dropna()
                .values
            )

        data_cut = data_pix[(data_pix > -window / 2) & (data_pix < window / 2)]

        if not data_cut.any().any():
            continue
        counts, bin_edges = np.histogram(
            data_cut,
            bins=(
                np.arange(np.min(data_cut), np.max(data_cut), step * 17.857)
            ),
        )
        if len(counts) < 1:
            continue
        bin_centers = (bin_edges - step * 17.857 / 2)[1:]

        try:
            params, _ = curve_fit(
                utils.gaussian,
                bin_centers,
                counts,
                p0=[
                    np.max(counts),
                    bin_centers[np.argmax(counts)],
                    100,
                    np.median(counts),
                ],
            )
        except (RuntimeError, TypeError):
            continue

        if i < 5:
            axes[0, i].plot(bin_centers, counts, ".", color="rebeccapurple")
            axes[0, i].plot(
                bin_centers,
                utils.gaussian(bin_centers, *params),
                "--",
                color="darkorange",
            )
            axes[0, i].set_xticks([])
            axes[0, i].set_yticks([])
            axes[0, i].set_title(f"{pixels[0]},{pix}")
        elif i >= 5 and i < 10:
            axes[1, i % 5].plot(
                bin_centers, counts, ".", color="rebeccapurple"
            )
            axes[1, i % 5].plot(
                bin_centers,
                utils.gaussian(bin_centers, *params),
                "--",
                color="darkorange",
            )
            axes[1, i % 5].set_xticks([])
            axes[1, i % 5].set_yticks([])
            axes[1, i % 5].set_title(f"{pixels[0]},{pix}")
        elif i >= 10 and i < 15:
            axes[2, i % 5].plot(
                bin_centers, counts, ".", color="rebeccapurple"
            )
            axes[2, i % 5].plot(
                bin_centers,
                utils.gaussian(bin_centers, *params),
                "--",
                color="darkorange",
            )
            axes[2, i % 5].set_xticks([])
            axes[2, i % 5].set_yticks([])
            axes[2, i % 5].set_title(f"{pixels[0]},{pix}")
        else:
            axes[3, i % 5].plot(
                bin_centers, counts, ".", color="rebeccapurple"
            )
            axes[3, i % 5].plot(
                bin_centers,
                utils.gaussian(bin_centers, *params),
                "--",
                color="darkorange",
            )
            axes[3, i % 5].set_xticks([])
            axes[3, i % 5].set_yticks([])
            axes[3, i % 5].set_title(f"{pixels[0]},{pix}")

    axes[1, 2].set_xlabel("\u0394t (ps)", fontsize=26)
    fig.text(0, 0.25, "# of coincidences (-)", fontsize=26, rotation=90)

    # Make plots tight
    plt.tight_layout()

    try:
        os.chdir(os.path.join(path, "results/ct_fit"))
    except FileNotFoundError:
        os.makedirs(os.path.join(path, "results/ct_fit"))
        os.chdir(os.path.join(path, "results/ct_fit"))
    plt.savefig(f"CT_fit_pixels_{pixels[0]},{pix}_grid.png")
    os.chdir(os.path.join(path, "cross_talk_data"))

    if show_plots:
        plt.show()
    else:
        plt.close(fig)


def collect_dcr_by_file(
    path: str,
    daughterboard_number: str,
    motherboard_number: str,
    firmware_version: str,
    timestamps: int = 1000,
):
    """Calculate dark count rate in counts per second per pixel.

    Calculate dark count rate for the given daughterboard and
    motherboard in units of counts per second per pixel for each file
    in the given folder and save the resulting list as a .pkl file.

    Parameters
    ----------
    path : str
        Path to the folder with the '.dat' data files.
    daughterboard_number : str
        LinoSPAD2 daughterboard number.
    motherboard_number : str
        LinoSPAD2 motherboard (FPGA) number, including the '#'.
    firmware_version : str
        LinoSPAD2 firmware version.
    timestamps : int, optional
        Number of timestamps per acquisition cycle per TDC. The default
        is 1000.

    Raises
    ------
    TypeError
        Raised if the firmware version given is not recognized.
    TypeError
        Raised if the daughterboard number given is not recognized.
    TypeError
        Raised if the motherboard number given is not recognized.
    """

    # Parameter type check
    if not isinstance(firmware_version, str):
        raise TypeError(
            "'firmware_version' should be a string, '2212b' or '2212s'"
        )
    if not isinstance(daughterboard_number, str):
        raise TypeError(
            "'daughterboard_number' should be a string, 'NL11' or 'A5'"
        )
    if not isinstance(motherboard_number, str):
        raise TypeError("'motherboard_number' should be a string")

    os.chdir(path)

    files = glob.glob("*.dat*")
    files = sorted(files)

    output_file_name = files[0][:-4] + "-" + files[-1][:-4]

    valid_per_pixel = np.zeros(256)

    dcr = []

    for i in tqdm(range(len(files)), desc="Going through files"):
        # Define matrix of pixel coordinates, where rows are numbers of TDCs
        # and columns are the pixels that connected to these TDCs
        if firmware_version == "2212s":
            pix_coor = np.arange(256).reshape(4, 64).T
        elif firmware_version == "2212b":
            pix_coor = np.arange(256).reshape(64, 4)
        else:
            print("\nFirmware version is not recognized.")
            sys.exit()

        # Unpack the data; offset calibration is not necessary
        data = f_up.unpack_binary_data(
            files[i],
            daughterboard_number,
            motherboard_number,
            firmware_version,
            timestamps,
            include_offset=False,
        )

        # Collect number of timestamps in each pixel - DCR
        for i in range(256):
            tdc, pix = np.argwhere(pix_coor == i)[0]
            ind = np.where(data[tdc].T[0] == pix)[0]
            ind1 = np.where(data[tdc].T[1][ind] > 0)[0]
            valid_per_pixel[i] = len(data[tdc].T[1][ind[ind1]])

        acq_window_length = np.max(data[:].T[1]) * 1e-12
        number_of_cycles = len(np.where(data[0].T[0] == -2)[0])

        dcr.append(valid_per_pixel / acq_window_length / number_of_cycles)

    dcr = np.array(dcr)

    # Save the results into a .pkl file
    try:
        os.chdir("dcr_data")
    except FileNotFoundError:
        os.mkdir("dcr_data")
        os.chdir("dcr_data")

    file_with_dcr = f"{output_file_name}_dcr_data.pkl"

    with open(file_with_dcr, "wb") as f:
        pickle.dump(dcr, f)

    # Check, if the file was created
    absolute_address = os.path.join(path, "dcr_data", file_with_dcr)
    if os.path.isfile(absolute_address) is True:
        print(f"\n> > > DCR data are saved in {absolute_address} < < <")
    else:
        print("File wasn't generated. Check input parameters.")

    # Average and median DCR, including the hot pixels
    dcr_average = np.average(dcr)
    dcr_median = np.median(dcr)

    # Average and median DCR, with the hot pixels masked
    mask = utils.apply_mask(daughterboard_number, motherboard_number)
    dcr[:, mask] = 0
    dcr_average_masked = np.average(dcr)
    dcr_median_masked = np.median(dcr)

    print(
        "\n".join(
            [
                "DCR including hot pixels:",
                f"DCR Average: {dcr_average:.0f} cps/pixel",
                f"DCR Median: {dcr_median:.0f} cps/pixel",
                "                                                 ",
                "DCR without hot pixels:",
                f"DCR Average: {dcr_average_masked:.0f} cps/pixel",
                f"DCR Median: {dcr_median_masked:.0f} cps/pixel",
            ]
        )
    )


def plot_dcr_histogram_and_stability(
    path: str,
    hist_number_of_bins: int = 200,
):
    """Plot median DCR vs file and histogram of DCR with integral.

    Plot median DCR vs file for checking the setup stability. Also, plot
    a histogram of DCR together with intergral over pixels in units of
    %.

    Parameters
    ----------
    path : str
        Path to the folder with '.dat' data files.
    hist_number_of_bins : int, optional
        Number of bins for the DCR histogram. The default is 200.

    Raises
    ------
    FileNotFoundError
        Raised if the folder with the DCR data was not found.
    FileNotFoundError
        Raised if the .pkl file with the DCR data was not found.
    """

    os.chdir(path)

    # Collect all files in the given folder
    files = glob.glob("*.dat")
    files = sorted(files)

    # Find the file with the DCR data for the found files
    dcr_file_name = files[0][:-4] + "-" + files[-1][:-4]
    file_with_dcr = f"{dcr_file_name}_dcr_data.pkl"

    try:
        os.chdir("dcr_data")
    except FileNotFoundError:
        raise FileNotFoundError("The folder with DCR data was not found.")

    try:
        with open(file_with_dcr, "rb") as f:
            data = pickle.load(f)
    except FileNotFoundError:
        raise FileNotFoundError(f"{file_with_dcr} was not found")

    # Median DCR over the whole sensor
    dcr_median = np.median(data)

    # Plot the DCR stability graph: median DCR vs file
    plt.rcParams.update({"font.size": 27})
    plt.figure(figsize=(16, 10))
    plt.plot(
        [x + 1 for x in range(len(data))],
        np.median(data, axis=1),
        color="darkslateblue",
        label=f"Median DCR: {dcr_median:.0f} cps/pixel",
    )
    plt.title("DCR stability")
    plt.xlabel("File (-)")
    plt.ylabel("Median DCR (cps)")
    plt.legend(loc="best")

    # Save the plot to "results/dcr"
    try:
        os.chdir("../results/dcr")
    except FileNotFoundError:
        os.makedirs("../results/dcr")
        os.chdir("../results/dcr")
    plt.savefig("DCR_stability_graph.png")

    # DCR histogram with integral
    # Compute the histogram
    bins = np.logspace(
        np.log10(0.1), np.log10(np.max(data[0])), hist_number_of_bins
    )
    hist, bin_edges = np.histogram(np.average(data, axis=0), bins=bins)
    bin_centers = (bin_edges - (bin_edges[1] - bin_edges[0]) / 2)[1:]

    # Plot the histogram
    fig, ax = plt.subplots(figsize=(16, 10))
    ax.bar(
        bin_centers,
        hist,
        width=np.diff(bin_edges),
        color="rebeccapurple",
    )

    # Calculate and plot the integral
    cumul = np.cumsum(hist)
    ax1 = ax.twinx()
    ax1.plot(bin_centers, cumul / 256 * 100, color="darkorange", linewidth=3)
    ax.set_xlim(10)
    ax1.set_xlim(10)
    ax.set_ylim(0)
    ax1.set_ylim(0)
    ax.set_xscale("log")
    ax.set_xlabel("DCR (cps/pixel)")
    ax.set_ylabel("Count (-)")
    ax1.set_ylabel("Integral (%)")
    ax.set_title(f"Median DCR: {dcr_median:.0f} cps/pixel")
    plt.show()

    # Save the plot to "results/dcr"
    fig.savefig("DCR_histogram_w_integral.png")


def _calculate_and_plot_cross_talk(
    path: str,
    pixels: List,
    multiplier: int,
    delta_window: float,
    senpop,
    pix_on_left: bool = False,
    feather_file_name: str = "",
    show_plots: bool = False,
):
    """Plot cross-talk peaks and calculate peak population.

    Plots a histogram of timestamp differences collected from the
    cross-talk data, fits using Gaussian to extract sigma and peak
    position. These two numbers are then used to calculate the peak
    population as numbers of data points in a +/- 2*sigma window
    around the peak position minus background in the same window.

    Parameters
    ----------
    path : str
        Path to the folder with the '.dat' data files.
    pixels : List[int]
        List of lists with pixels numbers. Each list starts with the
        aggressor pixel plus 20 pixels either on the left or right.
    multiplier : int
        Multiplier of the LinoSPAD2 average timing bin of 17.857 ps for
        histograms of the timestamp differences.
    delta_window : float
        Length of the window for the histogram range.
    senpop : list
        List with numbers of timestamps per pixel.
    pix_on_left : bool, optional
        Indicator of pixels to the left of the aggressor. The default is False.
    feather_file_name : str, optional
        Naming pattern of the feather files that can be used when
        actual data files are not available. The pattern should
        include the number of the first and the last data files, e.g.,
        "0000010000-0000012000". The default is "".
    show_plots : bool, optional
        Switch for showing the plots at the end. The default is False.

    Returns
    -------
    dict, dict
        Two dictionaries with distances in pixels from the aggressor
        pixel, where the first dictionary contains cross-talk
        probabilities and the second - the corresponding errors.
    """

    ct = []
    ct_err = []

    for pix in pixels:
        CT, CT_err = _plot_cross_talk_peaks(
            path,
            pixels=pix,
            multiplier=multiplier,
            window=delta_window,
            senpop=senpop,
            pix_on_left=pix_on_left,
            feather_file_name=feather_file_name,
            show_plots=show_plots,
        )
        _plot_cross_talk_grid(
            path,
            pixels=pix,
            multiplier=multiplier,
            window=delta_window,
            senpop=senpop,
            pix_on_left=pix_on_left,
            feather_file_name=feather_file_name,
            show_plots=show_plots,
        )

        ct.append(CT)
        ct_err.append(CT_err)

    return ct, ct_err


def _plot_cross_talk_vs_distance(
    path, ct, ct_err, pix_on_left: bool = False, show_plots: bool = False
):
    """Plot cross-talk vs distance from the aggressor.

    Plot cross-talk probability as a function of distance from the
    aggressor pixel.

    Parameters
    ----------
    path : str
        Path to the folder with the '.dat' data files.
    ct : dict
        Dictionary of cross-talk probabilities with distance from the
        aggressor as keys.
    ct_err : dict
        Dictionary of cross-talk error values with distance from the
        aggressor as keys.
    pix_on_left : bool, optional
        Indicator of pixels to the left of the aggressor. The default is False.
    show_plots : bool, optional
        Switch for showing the plots at the end. The default is False.
    """

    try:
        os.chdir(os.path.join(path, "ct_vs_distance"))
    except FileNotFoundError:
        os.makedirs(os.path.join(path, "ct_vs_distance"))
        os.chdir(os.path.join(path, "ct_vs_distance"))

    for _, (CT, CT_err) in enumerate(zip(ct, ct_err)):
        if CT == {} or CT_err == {}:
            continue

        differences = []
        keys = CT.keys()

        for key in keys:
            # Convert the key to a tuple
            key_tuple = eval(key)
            # Extract the difference and append it to the list
            differences.append(key_tuple[1] - key_tuple[0])

        fig = plt.figure(figsize=(16, 10))
        plt.rcParams.update({"font.size": 27})
        plt.errorbar(
            differences,
            list(CT.values()),
            list(CT_err.values()),
            fmt=".",
            color="indianred",
        )
        aggressor_pix = int(list(CT.keys())[0].split(",")[0].split("(")[1])
        plt.title(
            f"Cross-talk probability for aggressor pixel {aggressor_pix}"
        )
        plt.xlabel("Distance in pixels (-)")
        plt.ylabel("Cross-talk probability (%)")
        if pix_on_left:
            plt.savefig(
                f"Cross-talk_aggressor_pixel_{aggressor_pix}_onleft.png"
            )
            with open(
                f"Cross-talk_aggressor_pixel_{aggressor_pix}_onleft.pkl", "wb"
            ) as f:
                pickle.dump(fig, f)
        else:
            plt.savefig(
                f"Cross-talk_aggressor_pixel_{aggressor_pix}_onright.png"
            )
            with open(
                f"Cross-talk_aggressor_pixel_{aggressor_pix}_onright.pkl", "wb"
            ) as f:
                pickle.dump(fig, f)

        if show_plots:
            plt.show()
        else:
            plt.close(fig)


def _plot_average_cross_talk_vs_distance(
    path, ct, ct_err, pix_on_left: bool = False
):
    """Plot average cross-talk vs distance from the aggressor pixel.

    Plot average cross-talk probability vs distance from the aggressor
    averaging accross the whole sensor and all hot pixels byt dividing
    between either to the left or to the right of the aggressor the
    cross-talk is calculated.

    Parameters
    ----------
    path : str
        Path to the folder with the '.dat' data files.
    ct : dict
        Dictionary of cross-talk probabilities for each distance from 1 to 20.
    ct_err : dict
        Dictionary of cross-talk error values for each distance from 1 to 20.
    pix_on_left : bool, optional
        Indicator of pixels to the left of the aggressor. The default is False.

    Returns
    -------
    dict
        Dictionary of average cross-talk numbers and error with distance
        from the aggressor as keys.
    """

    try:
        os.chdir(os.path.join(path, "ct_vs_distance"))
    except FileNotFoundError:
        os.makedirs(os.path.join(path, "ct_vs_distance"))
        os.chdir(os.path.join(path, "ct_vs_distance"))

    if pix_on_left:
        final_result = {key: [] for key in range(-20, 0)}
        final_result_averages = {key: [] for key in range(-20, 0)}
    else:
        final_result = {key: [] for key in range(1, 21)}
        final_result_averages = {key: [] for key in range(1, 21)}

    for _, (ct_pick, ct_err_pick) in enumerate(zip(ct, ct_err)):
        if ct_pick == {} or ct_err_pick == {}:
            continue

        for key in ct_pick.keys():
            key_tuple = eval(key)

            key_difference = key_tuple[1] - key_tuple[0]
            final_result[key_difference].append(
                (ct_pick[key], ct_err_pick[key])
            )

    for key in final_result.keys():
        value = np.average([x[0] for x in final_result[key]])
        error = np.sqrt(np.sum([x[1] ** 2 for x in final_result[key]])) / len(
            final_result[key]
        )
        final_result_averages[key].append((value, error))

    fig = plt.figure(figsize=(16, 10))
    plt.rcParams.update({"font.size": 27})
    plt.title("Average cross-talk probability")
    plt.xlabel("Distance in pixels (-)")
    plt.ylabel("Cross-talk probability (%)")
    plt.errorbar(
        final_result_averages.keys(),
        [x[0][0] for x in final_result_averages.values()],
        yerr=[x[0][1] for x in final_result_averages.values()],
        fmt=".",
        color="darkred",
    )
    if pix_on_left:
        plt.savefig("Average_cross-talk_onleft.png")
        with open("Average_cross-talk_onleft.pkl", "wb") as f:
            pickle.dump(fig, f)
    else:
        plt.savefig("Average_cross-talk_onright.png")
        with open("Average_cross-talk_onright.pkl", "wb") as f:
            pickle.dump(fig, f)

    return final_result_averages


def zero_to_cross_talk_collect(
    path: str,
    pixels: List[int],
    rewrite: bool,
    daughterboard_number: str,
    motherboard_number: str,
    firmware_version: str,
    timestamps: int,
    include_offset: bool = False,
    delta_window: float = 50e3,
    apply_calibration: bool = True,
    absolute_timestamps: bool = False,
    correct_pix_address: bool = False,
):
    """Collect timestamp differences from cross-talk data.

    For the given list of hot pixels, collect all timestamp differences
    for pairs of pixels as aggressor-victim, where the victim pixel is
    taken from each side of the aggressor (where applicable) and up to
    the distance of 20 pixels (e.g., 100,120).

    Parameters
    ----------
    path : str
        Path to the folder with the '.dat' data files.
    pixels : list
        List of pixels to use as aggressors. These can be hot pixels
        or pixels where the signal is focused.
    rewrite : bool
        Switch for rewriting the feather file with timestamp differences.
        Used for conscious start of an hours-long data processing and
        avoiding rewriting/deleting already existing files.
    daughterboard_number : str
        LinoSPAD2 daughterboard number.
    motherboard_number : str
        LinoSPAD2 motherboard (FPGA) number, including the '#'.
    firmware_version : str
        LinoSPAD2 firmware version.
    timestamps : int
        Number of timestamps per TDC per cycle used during data
        collection.
    include_offset : bool
        Switch for including the offset calibration. The default is
        False.
    delta_window : float, optional
        Window size for collecting timestamp differences. The default
        is 50e3.
    apply_calibration : bool, optional
        Switch for applying callibration. The default is True.
    absolute_timestamps : bool, optional
        Indicator of data collected with absolute timestamps. The
        default is False.
    correct_pix_address : bool, optional
        Correct pixel address for the FPGA board on side 23. Here
        used to reverse the correction. The default is False.
    """

    print("\n> > > Collecting cross-talk data < < <\n")

    # Reverse correction if the motherboard connected to side "23" of the
    # daughterboard
    if correct_pix_address:
        for i, pixel in enumerate(pixels):
            if pixel > 127:
                pixels[i] = 255 - pixel
            else:
                pixels[i] = pixel + 128

    pixels_plus_20 = [
        [x + i for i in range(0, 21)] for x in pixels if x <= 235
    ]
    pixels_minus_20 = [
        [x - i for i in range(0, 21)] for x in pixels if x >= 20
    ]

    # Collecting sensor population
    os.chdir(path)
    files = glob.glob("*.dat")
    sensor_plot.collect_data_and_apply_mask(
        files,
        daughterboard_number,
        motherboard_number,
        firmware_version,
        timestamps,
        include_offset,
        apply_calibration,
        app_mask=False,
        save_to_file=True,
        absolute_timestamps=absolute_timestamps,
        correct_pix_address=correct_pix_address,
    )

    for pixels in pixels_plus_20:
        print(
            "Calculating timestamp differences between aggressor pixel "
            f"{pixels[0]} and pixels to the right"
        )
        _collect_cross_talk(
            path,
            pixels,
            rewrite,
            daughterboard_number,
            motherboard_number,
            firmware_version,
            timestamps,
            delta_window,
            include_offset,
            apply_calibration,
            absolute_timestamps,
            correct_pix_address,
        )

    for pixels in pixels_minus_20:
        print(
            "Calculating timestamp differences between aggressor pixel "
            f"{pixels[0]} and pixels to the left"
        )
        _collect_cross_talk(
            path,
            pixels,
            rewrite,
            daughterboard_number,
            motherboard_number,
            firmware_version,
            timestamps,
            delta_window,
            include_offset,
            apply_calibration,
            absolute_timestamps,
            correct_pix_address,
        )


def zero_to_cross_talk_plot(
    path: str,
    pixels: List[int],
    delta_window: float = 50e3,
    step: int = 10,
    show_plots: bool = False,
    feather_file_name: str = "",
):
    """Plot cross-talk peaks and as average vs distance from aggressor.

    For the given pixels, look for the timestamp differences in
    feather files, plot histograms and fits with Gaussian of cross-talk
    peaks for each pair of pixels (aggressor-victim), collects
    cross-talk peak population in a 2*sigma window, plots as average
    vs distance from the aggressor for each aggressor pixel and to each
    side (to the left and to the right). Finally, averages all numbers
    and produces a single plot of average cross-talk vs distance from
    the aggressor across all hot pixels.

    Parameters
    ----------
    path : str
        Path to the folder with the '.dat' data files.
    pixels : List[int]
        List of pixels to use as aggressors. These can be hot pixels
        or pixels where the signal is focused.
    delta_window : float, optional
        Window size for histograms of timestamp differences. The default
        is 50e3.
    step : int, optional
        Multiplier of average LinoSPAD2 time bin, which is 17.857 ps,
        for histograms. The default is 10.
    show_plots : bool, optional
        Switch for showing the plots at the end. The default is False.
    feather_file_name : str, optional
        Naming pattern of the feather files that can be used when
        actual data files are not available. The pattern should
        include the number of the first and the last data files, e.g.,
        "0000010000-0000012000". The default is "".

    Raises
    ------
    FileNotFoundError
        Raised when no txt file with the sensor population data is
        found.
    """

    print("\n> > > Plotting cross-talk peaks and averages < < <\n")
    try:
        os.chdir(os.path.join(path, "ct_vs_distance"))
    except FileNotFoundError:
        os.makedirs(os.path.join(path, "ct_vs_distance"))
        os.chdir(os.path.join(path, "ct_vs_distance"))

    pixels_plus_20 = [
        [x + i for i in range(0, 21)] for x in pixels if x <= 235
    ]
    pixels_minus_20 = [
        [x - i for i in range(0, 21)] for x in pixels if x >= 20
    ]
    try:
        os.chdir(os.path.join(path, "senpop_data"))
        senpop_data_txt = glob.glob("*.txt")[0]
        senpop = np.genfromtxt(senpop_data_txt)
    except Exception as _:
        raise FileNotFoundError(
            "Txt file with sensor population data is not found. Collect "
            "sensor population first."
        )

    ct_right, ct_err_right = _calculate_and_plot_cross_talk(
        path,
        pixels_plus_20,
        step,
        delta_window,
        senpop,
        pix_on_left=False,
        feather_file_name=feather_file_name,
        show_plots=show_plots,
    )
    ct_left, ct_err_left = _calculate_and_plot_cross_talk(
        path,
        pixels_minus_20,
        step,
        delta_window,
        senpop,
        pix_on_left=True,
        feather_file_name=feather_file_name,
        show_plots=show_plots,
    )

    _plot_cross_talk_vs_distance(
        path,
        ct_right,
        ct_err_right,
        pix_on_left=False,
        show_plots=show_plots,
    )
    _plot_cross_talk_vs_distance(
        path,
        ct_left,
        ct_err_left,
        pix_on_left=True,
        show_plots=show_plots,
    )

    averages_right = _plot_average_cross_talk_vs_distance(
        path,
        ct_right,
        ct_err_right,
        pix_on_left=False,
    )
    averages_left = _plot_average_cross_talk_vs_distance(
        path,
        ct_left,
        ct_err_left,
        pix_on_left=True,
    )

    on_both_average = {key: [] for key in range(1, 21)}
    for key in averages_left.keys():
        ct_value_average = (
            averages_left[key][0][0] + averages_right[np.abs(key)][0][0]
        ) / 2
        ct_error_average = (
            np.sqrt(
                averages_left[key][0][1] ** 2
                + averages_right[np.abs(key)][0][1] ** 2
            )
            / 2
        )
        on_both_average[np.abs(key)] = (ct_value_average, ct_error_average)

    fig = plt.figure(figsize=(16, 10))
    plt.rcParams.update({"font.size": 27})
    plt.title("Average cross-talk probability")
    plt.xlabel("Distance in pixels (-)")
    plt.ylabel("Cross-talk probability (%)")
    plt.yscale("log")
    plt.errorbar(
        on_both_average.keys(),
        [x[0] for x in on_both_average.values()],
        yerr=[x[1] for x in on_both_average.values()],
        fmt=".",
        color="darkred",
        label=f"Immediate neighbor: {on_both_average[1][0]:.2f}%",
    )
    plt.tight_layout()
    os.chdir(os.path.join(path, "ct_vs_distance"))
    plt.legend(loc="best")
    plt.savefig("Average_cross-talk.png")
    with open("Average_cross-talk.pkl", "wb") as f:
        pickle.dump(fig, f)

    return on_both_average, ct_right, ct_left



"""Module with scripts for unpacking data from LinoSPAD2.

This file can also be imported as a module and contains the following
functions:

    * unpack_binary_data - function for unpacking data from LinoSPAD2,
    firmware version 2212. Utilizes the numpy library to speed up the
    process.

    * unpack_binary_data_with_absolute_timestamps - function for
    unpacking data from LinoSPAD2, including the absolute timestamps,
    works with firmware version 2212.

"""

import os

import numpy as np

from LinoSPAD2.functions.calibrate import load_calibration_data


def unpack_binary_data(
    file: str,
    daughterboard_number: str,
    motherboard_number: str,
    firmware_version: str,
    timestamps: int = 512,
    include_offset: bool = False,
    apply_calibration: bool = True,
) -> np.ndarray:
    """Unpacks binary-encoded data from LinoSPAD2 firmware version 2212.

    Parameters
    ----------
    file : str
        Path to the binary data file.
    daughterboard_number : str
        LinoSPAD2 daughterboard number.
    motherboard_number : str
        LinoSPAD2 motherboard (FPGA) number, including the "#".
    firmware_version : str
        LinoSPAD2 firmware version. Either '2212s' (skip) or '2212b' (block).
    timestamps : int, optional
        Number of timestamps per cycle per TDC per acquisition cycle.
        The default is 512.
    include_offset : bool, optional
        Switch for applying offset calibration. The default is True.
    apply_calibration : bool, optional
        Switch for applying TDC and offset calibration. If set to 'True'
        while include_offset is set to 'False', only the TDC
        calibration is applied. The default is True.

    Returns
    -------
    data_all : array-like
        3D array of pixel coordinates in the TDC and the timestamps.

    Raises
    ------
    TypeError
        If 'daughterboard_number', 'motherboard_number', or 'firmware_version'
        parameters are not of string type.
    FileNotFoundError
        If no calibration data file is found.

    Notes
    -----
    The returned data is a 3D array where rows represent TDC numbers,
    columns represent the data, and each cell contains a pixel number in
    the TDC (from 0 to 3) and the timestamp recorded by that pixel.
    """
    # Parameter type check
    if not isinstance(daughterboard_number, str):
        raise TypeError("'daughterboard_number' should be a string.")
    if not isinstance(motherboard_number, str):
        raise TypeError("'motherboard_number' should be a string.")
    if not isinstance(firmware_version, str):
        raise TypeError("'firmware_version' should be a string.")

    # Unpack binary data
    raw_data = np.fromfile(file, dtype=np.uint32)
    # Timestamps are stored in the lower 28 bits
    data_timestamps = (raw_data & 0xFFFFFFF).astype(np.int64)
    # Pixel address in the given TDC is 2 bits above timestamp
    data_pixels = ((raw_data >> 28) & 0x3).astype(np.int8)
    # Check the top bit, assign '-1' to invalid timestamps
    data_timestamps[raw_data < 0x80000000] = -1
    # Free up memory
    del raw_data

    # Number of acquisition cycles in each data file
    cycles = len(data_timestamps) // (timestamps * 65)
    # Transform into a matrix of size 65 by cycles*timestamps
    data_pixels = (
        data_pixels.reshape(cycles, 65, timestamps)
        .transpose((1, 0, 2))
        .reshape(65, -1)
    )

    data_timestamps = (
        data_timestamps.reshape(cycles, 65, timestamps)
        .transpose((1, 0, 2))
        .reshape(65, -1)
    )

    # Cut the 65th TDC that does not hold any actual data from pixels
    data_pixels = data_pixels[:-1]
    data_timestamps = data_timestamps[:-1]

    # Insert '-2' at the end of each cycle
    insert_indices = np.linspace(
        timestamps, cycles * timestamps, cycles
    ).astype(np.int64)

    data_pixels = np.insert(
        data_pixels,
        insert_indices,
        -2,
        1,
    )
    data_timestamps = np.insert(
        data_timestamps,
        insert_indices,
        -2,
        1,
    )

    # Combine both matrices into a single one, where each cell holds pixel
    # coordinates in the TDC and the timestamp
    data_all = np.stack((data_pixels, data_timestamps), axis=2).astype(
        np.int64
    )

    if apply_calibration is False:
        data_all[:, :, 1] = data_all[:, :, 1] * 2500 / 140
    else:
        # Path to the calibration data
        pix_coordinates = np.arange(256).reshape(64, 4)

        path_calibration_data = os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            "..",
            "params",
            "calibration_data",
        )

        # Include the offset calibration or not
        try:
            if include_offset:
                calibration_matrix, offset_array = load_calibration_data(
                    path_calibration_data,
                    daughterboard_number,
                    motherboard_number,
                    firmware_version,
                    include_offset,
                )
            else:
                calibration_matrix = load_calibration_data(
                    path_calibration_data,
                    daughterboard_number,
                    motherboard_number,
                    firmware_version,
                    include_offset,
                )
        except FileNotFoundError:
            raise FileNotFoundError(
                "No .csv file with the calibration data was found. "
                "Check the path or run the calibration."
            )

        for i in range(256):
            # Transform pixel number to TDC number and pixel coordinates in
            # that TDC (from 0 to 3)
            tdc, pix = np.argwhere(pix_coordinates == i)[0]
            # Find data from that pixel
            ind = np.where(data_all[tdc].T[0] == pix)[0]
            # Cut non-valid timestamps ('-1's)
            ind = ind[data_all[tdc].T[1][ind] >= 0]
            if not np.any(ind):
                continue
            data_cut = data_all[tdc].T[1][ind]
            # Apply calibration; offset is added due to how delta ts are
            # calculated
            if include_offset:
                data_all[tdc].T[1][ind] = (
                    (data_cut - data_cut % 140) * 2500 / 140
                    + calibration_matrix[i, (data_cut % 140)]
                    + offset_array[i]
                )
            else:
                data_all[tdc].T[1][ind] = (
                    data_cut - data_cut % 140
                ) * 2500 / 140 + calibration_matrix[i, (data_cut % 140)]

    return data_all


def unpack_binary_data_with_absolute_timestamps(
    file_path: str,
    daughterboard_number: str,
    motherboard_number: str,
    firmware_version: str,
    timestamps: int = 512,
    include_offset: bool = False,
    apply_calibration: bool = True,
) -> np.ndarray:
    """Unpacks binary-encoded data from LinoSPAD2 firmware version 2212
    with absolute timestamps.

    Parameters
    ----------
    file_path : str
        Path to the binary data file.
    daughterboard_number : str
        LinoSPAD2 daughterboard number.
    motherboard_number : str
        LinoSPAD2 motherboard (FPGA) number, including the "#".
    firmware_version : str
        LinoSPAD2 firmware version. Either '2212s' (skip) or '2212b'
        (block) are accepted.
    timestamps : int, optional
        Number of timestamps per TDC per acquisition cycle.
        The default is 512.
    include_offset : bool, optional
        Switch for applying offset calibration. The default is True.
    apply_calibration : bool, optional
        Switch for applying TDC and offset calibration. If set to 'True'
        while include_offset is set to 'False', only the TDC
        calibration is applied. The default is True.

    Returns
    -------
    data_all : array-like
        3D array of pixel coordinates in the TDC and the timestamps.
    absolute_timestamps : array-like
        Array of absolute timestamps in ps, with length equal to the
        number of acquisition cycles, as the absolute timestamps are
        recorded at the start of each cycle.

    Raises
    ------
    TypeError
        If 'daughterboard_number', 'motherboard_number', or
        'firmware_version' parameters are not of string type.
    FileNotFoundError
        If no calibration data file is found.

    Notes
    -----
    The returned data is a 3D array where rows represent TDC numbers,
    columns represent the data, and each cell contains a pixel number in
    the TDC (from 0 to 3) and the timestamp recorded by that pixel.
    Additionally, an array of absolute timestamps is returned, representing
    the absolute time at the start of each acquisition cycle.

    Absolute timestamps are inserted at the start of each cycle. The
    value of these timestamps is given by the 400 MHz clock.
    """
    # Parameter type check
    if not isinstance(daughterboard_number, str):
        raise TypeError("'daughterboard_number' should be a string.")
    if not isinstance(motherboard_number, str):
        raise TypeError("'motherboard_number' should be a string.")
    if not isinstance(firmware_version, str):
        raise TypeError("'firmware_version' should be a string.")

    # Unpack binary data
    raw_data = np.fromfile(file_path, dtype=np.uint32)
    # Timestamps are stored in the lower 28 bits
    data_timestamps_all = (raw_data & 0xFFFFFFF).astype(np.int64)
    # Number of acquisition cycles in each data file
    cycles = len(data_timestamps_all) // (timestamps * 65 + 2)

    # List of indices of absolute timestamps
    ind = []
    # Converted absolute timestamps: from binary to ps
    absolute_timestamps = np.zeros(cycles)

    # Collect the absolute timestamps indices; absolute timestamps
    # inserted at the start of each cycle
    for cyc in range(cycles):
        ind.extend(
            [
                x
                for x in range(
                    cyc * (65 * timestamps + 2),
                    cyc * (65 * timestamps + 2) + 2,
                )
            ]
        )

    # Absolute timestamps only
    data_absolute_timestamps = data_timestamps_all[ind].reshape(cycles, 2)

    # Convert the absolute timestamps from binary to decimal (ps)
    for cyc in range(cycles):
        absolute_timestamps[cyc] = int(
            "0b"
            + bin(int(data_absolute_timestamps[cyc][1]))[2:]
            + bin(int(data_absolute_timestamps[cyc][0]))[2:],
            2,
        )

    del data_absolute_timestamps

    # Cut the absolute timestamps, collect the timestamps
    raw_data_cut = np.delete(raw_data, ind)
    data_timestamps_cut = (raw_data_cut & 0xFFFFFFF).astype(np.int64)
    data_timestamps_cut[raw_data_cut < 0x80000000] = -1
    # Pixel address in the given TDC is 2 bits above timestamp
    data_pixels = ((raw_data_cut >> 28) & 0x3).astype(np.int8)
    del raw_data
    # Standard data: everything besides the absolute timestamps

    # Transform into matrix 65 by cycles*timestamps
    data_matrix_pixels = (
        data_pixels.reshape(cycles, 65, timestamps)
        .transpose((1, 0, 2))
        .reshape(65, -1)
    )

    data_matrix_timestamps = (
        data_timestamps_cut.reshape(cycles, 65, timestamps)
        .transpose((1, 0, 2))
        .reshape(65, -1)
    )
    # Cut the 65th TDC that does not hold any actual data from pixels
    data_matrix_pixels = data_matrix_pixels[:-1]
    data_matrix_timestamps = data_matrix_timestamps[:-1]
    # Insert '-2' at the end of each cycle
    insert_indices = np.linspace(
        timestamps, cycles * timestamps, cycles
    ).astype(np.longlong)
    data_matrix_pixels = np.insert(
        data_matrix_pixels,
        insert_indices,
        -2,
        1,
    )

    data_matrix_timestamps = np.insert(
        data_matrix_timestamps,
        insert_indices,
        -2,
        1,
    )
    # Combine both matrices into a single one, where each cell holds pixel
    # coordinates in the TDC and the timestamp
    data_all = np.stack(
        (data_matrix_pixels, data_matrix_timestamps), axis=2
    ).astype(np.int64)

    if apply_calibration is False:
        data_all[:, :, 1] = data_all[:, :, 1] * 2500 / 140
    else:
        # Path to the calibration data
        pixel_coordinates = np.arange(256).reshape(64, 4)
        path_calibration_data = os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            "..",
            "params",
            "calibration_data",
        )

        # Include the offset calibration or not
        try:
            if include_offset:
                calibration_matrix, offset_array = load_calibration_data(
                    path_calibration_data,
                    daughterboard_number,
                    motherboard_number,
                    firmware_version,
                    include_offset,
                )
            else:
                calibration_matrix = load_calibration_data(
                    path_calibration_data,
                    daughterboard_number,
                    motherboard_number,
                    firmware_version,
                    include_offset,
                )
        except FileNotFoundError:
            raise FileNotFoundError(
                "No .csv file with the calibration data was found. "
                "Check the path or run the calibration."
            )

        for i in range(256):
            # Transform pixel number to TDC number and pixel coordinates in
            # that TDC (from 0 to 3)
            tdc, pix = np.argwhere(pixel_coordinates == i)[0]
            # Find data from that pixel
            ind = np.where(data_all[tdc].T[0] == pix)[0]
            # Cut non-valid timestamps ('-1's)
            ind = ind[np.where(data_all[tdc].T[1][ind] >= 0)[0]]
            if not np.any(ind):
                continue
            data_cut = data_all[tdc].T[1][ind]
            # Apply calibration; offset is added due to how delta ts are
            # calculated
            if include_offset:
                data_all[tdc].T[1][ind] = (
                    (data_cut - data_cut % 140) * 2500 / 140
                    + calibration_matrix[i, (data_cut % 140)]
                    + offset_array[i]
                )
            else:
                data_all[tdc].T[1][ind] = (
                    data_cut - data_cut % 140
                ) * 2500 / 140 + calibration_matrix[i, (data_cut % 140)]

    return data_all, absolute_timestamps



import os
import unittest

import numpy as np

from LinoSPAD2.functions.unpack import unpack_binary_data


class TestUnpackBin(unittest.TestCase):
    def test_valid_input(self):
        # Positive test case with valid inputs
        work_dir = r"{}".format(
            os.path.dirname(os.path.realpath(__file__)) + "/.."
        )
        os.chdir(work_dir)
        file = r"tests/test_data/test_data_2212b.dat"
        daughterboard_number = "NL11"
        motherboard_number = "#33"
        timestamps = 300
        firmware_version = "2212b"
        include_offset = False

        data_all = unpack_binary_data(
            file,
            daughterboard_number,
            motherboard_number,
            firmware_version,
            timestamps,
            include_offset,
        )

        # Assert the shape of the output data
        self.assertEqual(data_all.shape, (64, 300 * 300 + 300, 2))
        # Assert the data type of the output data
        self.assertEqual(data_all.dtype, np.longlong)


if __name__ == "__main__":
    unittest.main()
